import { createSessionCookie } from "./cookie.js";
import { logError } from "../utils/log.js";
import { generateHashWithScrypt, generateRandomString, validateScryptHash } from "../utils/crypto.js";
import { LuciaError } from "./error.js";
import { parseCookie } from "../utils/cookie.js";
import { transformDatabaseSessionData } from "./session.js";
import { validateDatabaseKey, getOneTimeKeyExpiration } from "./key.js";
export { SESSION_COOKIE_NAME } from "./cookie.js";
export const lucia = (config) => {
    return new Auth(config);
};
const validateConfiguration = (config) => {
    const isAdapterProvided = config.adapter;
    if (!isAdapterProvided) {
        logError('Adapter is not defined in configuration ("config.adapter")');
        process.exit(1);
    }
};
export class Auth {
    adapter;
    generateUserId;
    sessionCookie;
    sessionTimeout;
    ENV;
    hash;
    autoDatabaseCleanup;
    transformUserData;
    csrfProtection;
    constructor(config) {
        validateConfiguration(config);
        const defaultSessionCookieOption = {
            sameSite: "lax",
            path: "/"
        };
        if ("user" in config.adapter) {
            if ("getSessionAndUserBySessionId" in config.adapter.user) {
                delete config.adapter.user.getSessionAndUserBySessionId;
            }
            if ("getSessionAndUserBySessionId" in config.adapter.session) {
                delete config.adapter.session.getSessionAndUserBySessionId;
            }
        }
        this.adapter =
            "user" in config.adapter
                ? {
                    ...config.adapter.user(LuciaError),
                    ...config.adapter.session(LuciaError)
                }
                : config.adapter(LuciaError);
        this.generateUserId =
            config.generateCustomUserId ?? (() => generateRandomString(15));
        this.ENV = config.env;
        this.csrfProtection = config.csrfProtection ?? true;
        this.sessionTimeout = {
            activePeriod: config.sessionTimeout?.activePeriod ?? 1000 * 60 * 60 * 24,
            idlePeriod: config.sessionTimeout?.idlePeriod ?? 1000 * 60 * 60 * 24 * 14
        };
        this.autoDatabaseCleanup = config.autoDatabaseCleanup ?? true;
        this.transformUserData = ({ id, hashed_password, provider_id, ...attributes }) => {
            const defaultTransform = ({ id }) => {
                return {
                    userId: id
                };
            };
            const transform = config.transformUserData ?? defaultTransform;
            return transform({ id, ...attributes });
        };
        this.sessionCookie = config.sessionCookie ?? [defaultSessionCookieOption];
        this.hash = {
            generate: config.hash?.generate ?? generateHashWithScrypt,
            validate: config.hash?.validate ?? validateScryptHash
        };
    }
    getUser = async (userId) => {
        const databaseUser = await this.adapter.getUser(userId);
        if (!databaseUser)
            throw new LuciaError("AUTH_INVALID_USER_ID");
        const user = this.transformUserData(databaseUser);
        return user;
    };
    getSessionUser = async (sessionId) => {
        if (sessionId.length !== 40)
            throw new LuciaError("AUTH_INVALID_SESSION_ID");
        let userData;
        let sessionData;
        if (this.adapter.getSessionAndUserBySessionId !== undefined) {
            const databaseUserSession = await this.adapter.getSessionAndUserBySessionId(sessionId);
            if (!databaseUserSession)
                throw new LuciaError("AUTH_INVALID_SESSION_ID");
            userData = databaseUserSession.user;
            sessionData = databaseUserSession.session;
        }
        else {
            sessionData = await this.adapter.getSession(sessionId);
            userData = sessionData
                ? await this.adapter.getUser(sessionData.user_id)
                : null;
        }
        if (!sessionData)
            throw new LuciaError("AUTH_INVALID_SESSION_ID");
        const session = transformDatabaseSessionData(sessionData);
        if (!session) {
            if (this.autoDatabaseCleanup) {
                await this.adapter.deleteSession(sessionId);
            }
            throw new LuciaError("AUTH_INVALID_SESSION_ID");
        }
        if (!userData)
            throw new LuciaError("AUTH_INVALID_USER_ID");
        return {
            user: this.transformUserData(userData),
            session
        };
    };
    createUser = async (data) => {
        const userId = await this.generateUserId();
        const userAttributes = data.attributes ?? {};
        if (data.key === null) {
            const userData = await this.adapter.setUser(userId, userAttributes, null);
            const user = this.transformUserData(userData);
            return user;
        }
        const keyId = `${data.key.providerId}:${data.key.providerUserId}`;
        const password = data.key.password;
        const hashedPassword = password ? await this.hash.generate(password) : null;
        const userData = await this.adapter.setUser(userId, userAttributes, {
            id: keyId,
            user_id: userId,
            hashed_password: hashedPassword,
            primary: true,
            expires: null
        });
        const user = this.transformUserData(userData);
        return user;
    };
    updateUserAttributes = async (userId, attributes) => {
        const [userData] = await Promise.all([
            this.adapter.updateUserAttributes(userId, attributes),
            this.autoDatabaseCleanup
                ? await this.deleteDeadUserSessions(userId)
                : null
        ]);
        const user = this.transformUserData(userData);
        return user;
    };
    deleteUser = async (userId) => {
        await this.adapter.deleteSessionsByUserId(userId);
        await this.adapter.deleteKeysByUserId(userId);
        await this.adapter.deleteUser(userId);
    };
    validateKeyPassword = async (providerId, providerUserId, password) => {
        const keyId = `${providerId}:${providerUserId}`;
        const databaseKeyData = await this.adapter.getKey(keyId);
        if (!databaseKeyData)
            throw new LuciaError("AUTH_INVALID_KEY_ID");
        const hashedPassword = databaseKeyData.hashed_password;
        if (!hashedPassword)
            throw new LuciaError("AUTH_INVALID_PASSWORD");
        if (hashedPassword.startsWith("$2a"))
            throw new LuciaError("AUTH_OUTDATED_PASSWORD");
        const isValidPassword = await this.hash.validate(password, hashedPassword);
        if (!isValidPassword)
            throw new LuciaError("AUTH_INVALID_PASSWORD");
        const key = validateDatabaseKey(databaseKeyData);
        return key;
    };
    getSession = async (sessionId) => {
        if (sessionId.length !== 40)
            throw new LuciaError("AUTH_INVALID_SESSION_ID");
        const databaseSession = await this.adapter.getSession(sessionId);
        if (!databaseSession)
            throw new LuciaError("AUTH_INVALID_SESSION_ID");
        const session = transformDatabaseSessionData(databaseSession);
        if (!session) {
            if (this.autoDatabaseCleanup) {
                await this.adapter.deleteSession(sessionId);
            }
            throw new LuciaError("AUTH_INVALID_SESSION_ID");
        }
        return session;
    };
    getAllUserSessions = async (userId) => {
        await this.getUser(userId);
        const databaseData = await this.adapter.getSessionsByUserId(userId);
        const storedUserSessions = databaseData.map((val) => {
            const session = transformDatabaseSessionData(val);
            if (session) {
                return {
                    isValid: true,
                    session
                };
            }
            return {
                isValid: false,
                sessionId: val.id
            };
        });
        if (storedUserSessions.some((val) => val.isValid === false)) {
            await Promise.all(storedUserSessions
                .filter((val) => val.isValid === false)
                .map((val) => val));
        }
        const validUserSessions = storedUserSessions
            .map((val) => {
            return val.isValid ? val.session : null;
        })
            .filter((val) => val !== null);
        return validUserSessions;
    };
    validateSession = async (sessionId) => {
        const session = await this.getSession(sessionId);
        if (session.state === "active")
            return session;
        const renewedSession = await this.renewSession(sessionId);
        return renewedSession;
    };
    validateSessionUser = async (sessionId) => {
        const { session, user } = await this.getSessionUser(sessionId);
        if (session.state === "active")
            return { session, user };
        const renewedSession = await this.renewSession(sessionId);
        return {
            session: renewedSession,
            user
        };
    };
    generateSessionId = () => {
        const sessionId = generateRandomString(40);
        const activePeriodExpires = new Date(new Date().getTime() + this.sessionTimeout.activePeriod);
        const idlePeriodExpires = new Date(activePeriodExpires.getTime() + this.sessionTimeout.idlePeriod);
        return [sessionId, activePeriodExpires, idlePeriodExpires];
    };
    createSession = async (userId) => {
        const [sessionId, activePeriodExpires, idlePeriodExpires] = this.generateSessionId();
        await Promise.all([
            this.adapter.setSession({
                id: sessionId,
                user_id: userId,
                active_expires: activePeriodExpires.getTime(),
                idle_expires: idlePeriodExpires.getTime()
            }),
            this.autoDatabaseCleanup
                ? await this.deleteDeadUserSessions(userId)
                : null
        ]);
        return {
            userId,
            activePeriodExpires,
            sessionId,
            idlePeriodExpires,
            state: "active",
            isFresh: true
        };
    };
    renewSession = async (sessionId) => {
        if (sessionId.length !== 40)
            throw new LuciaError("AUTH_INVALID_SESSION_ID");
        const sessionData = await this.adapter.getSession(sessionId);
        if (!sessionData)
            throw new LuciaError("AUTH_INVALID_SESSION_ID");
        const session = transformDatabaseSessionData(sessionData);
        if (!session) {
            if (this.autoDatabaseCleanup) {
                await this.adapter.deleteSession(sessionId);
            }
            throw new LuciaError("AUTH_INVALID_SESSION_ID");
        }
        const [renewedSession] = await Promise.all([
            await this.createSession(session.userId),
            this.autoDatabaseCleanup
                ? await this.deleteDeadUserSessions(session.userId)
                : null
        ]);
        return renewedSession;
    };
    invalidateSession = async (sessionId) => {
        await this.adapter.deleteSession(sessionId);
    };
    invalidateAllUserSessions = async (userId) => {
        await this.adapter.deleteSessionsByUserId(userId);
    };
    deleteDeadUserSessions = async (userId) => {
        const sessions = await this.adapter.getSessionsByUserId(userId);
        const currentTime = new Date().getTime();
        const deadSessionIds = sessions
            .filter((val) => val.idle_expires < currentTime)
            .map((val) => val.id);
        if (deadSessionIds.length === 0)
            return;
        await this.adapter.deleteSession(...deadSessionIds);
    };
    validateRequestHeaders = (request) => {
        const cookies = parseCookie(request.headers.get("cookie") ?? "");
        const sessionId = cookies.auth_session ?? "";
        const checkForCsrf = request.method !== "GET" && request.method !== "HEAD";
        if (checkForCsrf && this.csrfProtection) {
            const origin = request.headers.get("Origin");
            if (!origin)
                throw new LuciaError("AUTH_INVALID_REQUEST");
            const url = new URL(request.url);
            if (url.origin !== origin)
                throw new LuciaError("AUTH_INVALID_REQUEST");
        }
        return sessionId;
    };
    createSessionCookies = (session) => {
        return this.sessionCookie.map((options) => createSessionCookie(session, this.ENV, options));
    };
    createKey = async (userId, keyData) => {
        const keyId = `${keyData.providerId}:${keyData.providerUserId}`;
        const hashedPassword = keyData.password
            ? await this.hash.generate(keyData.password)
            : null;
        const oneTimeExpires = getOneTimeKeyExpiration(keyData.timeout);
        await this.adapter.setKey({
            id: keyId,
            user_id: userId,
            hashed_password: hashedPassword,
            primary: false,
            expires: oneTimeExpires?.getTime() ?? null
        });
        return {
            providerId: keyData.providerId,
            providerUserId: keyData.providerUserId,
            isPrimary: false,
            isPasswordDefined: !!keyData.password,
            userId,
            oneTimeExpires: oneTimeExpires ?? null
        };
    };
    deleteKey = async (providerId, providerUserId) => {
        const keyId = `${providerId}:${providerUserId}`;
        await this.adapter.deleteNonPrimaryKey(keyId);
    };
    getKey = async (providerId, providerUserId) => {
        const keyId = `${providerId}:${providerUserId}`;
        const keyData = await this.adapter.getKey(keyId);
        if (!keyData)
            throw new LuciaError("AUTH_INVALID_KEY_ID");
        const key = validateDatabaseKey(keyData);
        return key;
    };
    getKeyUser = async (providerId, providerUserId) => {
        const key = await this.getKey(providerId, providerUserId);
        const user = await this.getUser(key.userId);
        return {
            key,
            user
        };
    };
    getAllUserKeys = async (userId) => {
        await this.getUser(userId);
        const databaseData = await this.adapter.getKeysByUserId(userId);
        return databaseData
            .map((val) => {
            try {
                return validateDatabaseKey(val);
            }
            catch {
                return null;
            }
        })
            .filter((maybeKey) => maybeKey !== null);
    };
    updateKeyPassword = async (providerId, providerUserId, password) => {
        const keyId = `${providerId}:${providerUserId}`;
        const hashedPassword = password ? await this.hash.generate(password) : null;
        await this.adapter.updateKeyPassword(keyId, hashedPassword);
    };
}
